/******************************************************************************
  This file contains routines that can be bound to a Postgres backend and
  called by the backend in the process of processing queries.  The calling
  format for these routines is dictated by Postgres architecture.
******************************************************************************/

#ifdef __cplusplus
extern "C" 
{
#endif
#include "postgres.h"
#include <ctype.h>
#include "graph.h"
#include "graph_op.h"
#ifdef __cplusplus
}
#endif


void GraphNodeSpecParser( UNPACKED_GRAPH *graph, char *str )
{
   char *p;

   if ( (p=strtok(str,", "))!=NULL )
   {
      UnpackedGraph_InsertNode(graph,p);

      while ( (p=strtok(NULL,", ")) != NULL )
      {
	   UnpackedGraph_InsertNode(graph,p);
      }
   }
}


void GraphEdgeSpecParser( UNPACKED_GRAPH *graph, char *str )
{
   char *offset, *p, *q;

   for(offset=str;;)
   {
      if ( (p=strchr(offset,'('))!=NULL )
      {
         if ( (q=strchr(p+1,')'))!=NULL )
         {
            *q = 0;
            offset = q + 1;
            UnpackedGraph_InsertEdge_BySpec( graph, p+1, 0 );
         }
         else  elog( ERROR, "( without a matching )" );
      }
      else break;
   }   
}

void GraphLabelSpecParser( UNPACKED_GRAPH *graph, char * str )
{
   char *offset, *p, *q;

   for(offset=str;;)
   {
      if ( (p=strchr(offset,'('))!=NULL )
      {
         if ( (q=strchr(p+1,')'))!=NULL )
         {
            *q = 0;
            offset = q+1;
	    UnpackedGraph_InsertLabel_BySpec( graph, p+1, 0 );
         }
         else  elog( ERROR, "( without a matching )" );
      }
      else break;
   }
}

UNPACKED_GRAPH *GraphSpecParser( char *spec )
{
   UNPACKED_GRAPH *result;
   char *components[3], *p, *q, *offset;
   int i;

   result = NULL;
   
   /* 
    *  we first separate the three components from the specification - node list, edge list, label list
    *  notice that the specification looks like:
    *  { node1, node2, ...., nodeN } { (from1,to1), (from2,to2), ... (fromN,toN) } 
    *  { (from1,to1,label1), (from2,to2,label2), ..., (fromN,toN,labelN) }
    */
   offset = spec;
   for(i=0;i<3;i++)
   {
      if ( (p=strchr(offset,'{'))!=NULL )
      {
         if ( (q=strchr(p+1,'}'))!=NULL )
         { 
            *q = '\0';
            components[i] = p+1;
            offset = q+1;
         }
         else elog( ERROR, "{ without a matching }" );
      }
      else elog( ERROR, "illegal graph specification" );
   }

   result = UnpackedGraph_Create();

   GraphNodeSpecParser( result, components[0] );
   GraphEdgeSpecParser( result, components[1] );
   GraphLabelSpecParser( result, components[2] );
   
   return result;
}

PG_FUNCTION_INFO_V1(graph_in);
Datum graph_in(PG_FUNCTION_ARGS);

PG_FUNCTION_INFO_V1(graph_out);
Datum graph_out(PG_FUNCTION_ARGS);

Datum graph_in(PG_FUNCTION_ARGS)
{
   char *str = (char *) PG_GETARG_POINTER(0);
   UNPACKED_GRAPH *ug;
   PACKED_GRAPH *pg;

   ug = GraphSpecParser(str);
   pg = Pack(ug);
   UnpackedGraph_Free(ug);

   PG_RETURN_POINTER(pg);
}

Datum graph_out(PG_FUNCTION_ARGS)
{
   PACKED_GRAPH *graph = PG_GETARG_GRAPH(0);
   char *outstr, *p;
   GRAPH_OFFSET *node_table, *label_table, *edges, *offset, from, to;
   char *nodes, *labels;   
   char buffer[1024];
   int i, capacity, length, len, j;
 
   capacity = 1024;
   length = 0;
   outstr = (char *) palloc(capacity);
   outstr[length] = 0;

   node_table = (GRAPH_OFFSET *) PACKED_GRAPH_NODE_TABLE(graph);
   label_table = (GRAPH_OFFSET *) PACKED_GRAPH_LABEL_TABLE(graph);
   nodes = (char *) PACKED_GRAPH_NODES(graph);
   edges = (GRAPH_OFFSET *) PACKED_GRAPH_EDGES(graph);
   labels = (char *) PACKED_GRAPH_LABELS(graph);

   outstr = AddString( outstr, &capacity, &length, "{ ", 0 );
   p = PACKED_GRAPH_NODES(graph);
   j = graph->nnodes-1;

   for(i=0;i<j;i++)
   {
      sprintf( buffer, "%s,", p);
      len = strlen(p)+1;
      outstr = AddString( outstr, &capacity, &length, buffer, 0 );
      p += len;
   }
   if ( graph->nnodes>0 )
   {
     sprintf( buffer, "%s", p );
     outstr = AddString( outstr, &capacity, &length, buffer, 0 );
   }

   outstr = AddString( outstr, &capacity, &length, " } { ", 0 );      

   offset = (GRAPH_OFFSET *) PACKED_GRAPH_EDGES(graph);

   j = graph->nedges-1;
   for(i=0;i<j;i++)
   {
      from = offset[0];
      to   = offset[1];
      sprintf( buffer, "(%s,%s),", nodes+node_table[from], nodes+node_table[to]);
      outstr = AddString( outstr, &capacity, &length, buffer, 0);
      offset += 2;
   }
   if ( graph->nedges>0 )
   {  
      from = offset[0];
      to   = offset[1];
      sprintf( buffer, "(%s,%s)",nodes+node_table[from], nodes+node_table[to]);
      outstr = AddString( outstr, &capacity, &length, buffer, 0 );
   }
   outstr = AddString( outstr, &capacity, &length, " } { ", 0 );

   j= graph->nlabels-1;
   for(i=0;i<j;i++)
   {
      p = (char *) (labels + label_table[i]);  
      from = (GRAPH_OFFSET) *p;                  
      to   = (GRAPH_OFFSET) *(p+sizeof(GRAPH_OFFSET));
      p += (2*sizeof(GRAPH_OFFSET));
      sprintf( buffer,"(%s,%s,%s),", nodes+node_table[from], nodes+node_table[to], p );
      outstr = AddString( outstr, &capacity, &length, buffer,0 );
   }
   if ( graph->nlabels>0 )
   {
      p = (char *) (labels + label_table[graph->nlabels-1]);
      from = (GRAPH_OFFSET) *p;
      to   = (GRAPH_OFFSET) *(p+sizeof(GRAPH_OFFSET));
      p += (2*sizeof(GRAPH_OFFSET));
      sprintf( buffer,"(%s,%s,%s) }", nodes+node_table[from], nodes+node_table[to], p );
      outstr = AddString( outstr, &capacity, &length, buffer,0 );
   }
   outstr = AddString( outstr, &capacity, &length, " }", 0 );

   PG_RETURN_POINTER(outstr);
}


// operators

PG_FUNCTION_INFO_V1(graph_node_count);
PG_FUNCTION_INFO_V1(graph_edge_count);
PG_FUNCTION_INFO_V1(graph_label_count);
PG_FUNCTION_INFO_V1(graph_indegree);
PG_FUNCTION_INFO_V1(graph_outdegree);
PG_FUNCTION_INFO_V1(graph_has_node);
PG_FUNCTION_INFO_V1(graph_has_edge);
PG_FUNCTION_INFO_V1(graph_has_edge_1);
PG_FUNCTION_INFO_V1(graph_add_node);
PG_FUNCTION_INFO_V1(graph_add_edge);
PG_FUNCTION_INFO_V1(graph_add_label);
PG_FUNCTION_INFO_V1(graph_add_graph);
PG_FUNCTION_INFO_V1(graph_breadth_first_search);
PG_FUNCTION_INFO_V1(graph_depth_first_search);
PG_FUNCTION_INFO_V1(graph_has_self_loop);
PG_FUNCTION_INFO_V1(graph_out_edges);
PG_FUNCTION_INFO_V1(graph_in_edges);
PG_FUNCTION_INFO_V1(graph_connected_components);


Datum graph_node_count(PG_FUNCTION_ARGS)
{
   PACKED_GRAPH *graph = PG_GETARG_GRAPH(0);
   PG_RETURN_INT32(graph->nnodes);
}

Datum graph_edge_count(PG_FUNCTION_ARGS)
{
   PACKED_GRAPH *graph = PG_GETARG_GRAPH(0);
   PG_RETURN_INT32(graph->nedges);
}

Datum graph_label_count(PG_FUNCTION_ARGS)
{
   PACKED_GRAPH *graph = PG_GETARG_GRAPH(0);
   PG_RETURN_INT32(graph->nlabels);
}

Datum graph_indegree(PG_FUNCTION_ARGS)
{
   PACKED_GRAPH *graph = PG_GETARG_GRAPH(0);
   char  *name  = GET_STR(PG_GETARG_TEXT_P(1));
   GRAPH_OFFSET *offset;
   char  *nodes;
   int i, j, count;

   count = 0;
   nodes = PACKED_GRAPH_NODES(graph);
   offset = (GRAPH_OFFSET *)PACKED_GRAPH_NODE_TABLE(graph);

   for(i=0;i<graph->nnodes;i++)
      if ( strcmp(nodes+offset[i], name)==0 ) break;

   if (i<graph->nnodes)
   {
      offset = (GRAPH_OFFSET *) PACKED_GRAPH_EDGES(graph);
      for(j=0;j<graph->nedges;j++)
      {
         if ( i==offset[1] )  count++;
         offset+=2;
      }
   }
   
   PG_RETURN_INT32(count);

}

Datum graph_outdegree(PG_FUNCTION_ARGS)
{
   PACKED_GRAPH *graph = PG_GETARG_GRAPH(0);
   char  *name  = GET_STR(PG_GETARG_TEXT_P(1));
   GRAPH_OFFSET *offset;
   char  *nodes;
   int i, j, count;

   count = 0;
   nodes = PACKED_GRAPH_NODES(graph);
   offset = (GRAPH_OFFSET *)PACKED_GRAPH_NODE_TABLE(graph);

   for(i=0;i<graph->nnodes;i++)
      if ( strcmp(nodes+offset[i], name)==0 ) break;

   if (i<graph->nnodes)
   {
      offset = (GRAPH_OFFSET *) PACKED_GRAPH_EDGES(graph);
      for(j=0;j<graph->nedges;j++)
      {
         if ( i==*offset )  count++;
         offset+=2;
      }
   }

   PG_RETURN_INT32(count);

}


Datum graph_has_node(PG_FUNCTION_ARGS)
{
   PACKED_GRAPH *graph = PG_GETARG_GRAPH(0);
   char  *name  = GET_STR(PG_GETARG_TEXT_P(1));

   if ( PackedGraph_FindNode(graph,name)==-1 )  PG_RETURN_BOOL(FALSE);
   PG_RETURN_BOOL(TRUE);
}


Datum graph_has_edge_1(PG_FUNCTION_ARGS)
{
   PACKED_GRAPH *graph = PG_GETARG_GRAPH(0);
   char *spec = GET_STR(PG_GETARG_TEXT_P(1));
   char *src, *dest;
   int ret;

   src = strtok( spec, ", ");
   if ( src==NULL )  elog( ERROR, "syntax error for edge" );
   dest = strtok( NULL, ", ");
   if ( dest==NULL ) elog( ERROR, "syntax error for edge" );

   ret = PackedGraph_FindEdge_ByName(graph, src, dest );
   PG_RETURN_BOOL((ret!=-1));
}

Datum graph_has_edge(PG_FUNCTION_ARGS)
{
   PACKED_GRAPH *graph = PG_GETARG_GRAPH(0);
   char  *src   = GET_STR(PG_GETARG_TEXT_P(1));
   char  *dest  = GET_STR(PG_GETARG_TEXT_P(2));

   PG_RETURN_BOOL( (-1!=PackedGraph_FindEdge_ByName(graph,src,dest)));
}

Datum graph_add_node(PG_FUNCTION_ARGS)
{
   PACKED_GRAPH *graph = PG_GETARG_GRAPH(0);
   char *node = GET_STR(PG_GETARG_TEXT_P(1));

   graph = PackedGraph_InsertNode(graph,node);
   PG_RETURN_POINTER(graph);
}

Datum graph_add_edge(PG_FUNCTION_ARGS)
{
   PACKED_GRAPH *graph = PG_GETARG_GRAPH(0);
   char *edges = GET_STR(PG_GETARG_TEXT_P(1));

   graph = PackedGraph_InsertEdge(graph,edges);
   PG_RETURN_POINTER(graph);
}

Datum graph_add_label(PG_FUNCTION_ARGS)
{
   PACKED_GRAPH *graph = PG_GETARG_GRAPH(0);
   char *labels = GET_STR(PG_GETARG_TEXT_P(1));
   char *buffer;

   buffer = labels;

   graph = PackedGraph_InsertLabel(graph,buffer);
   PG_RETURN_POINTER(graph);
}

Datum graph_add_graph(PG_FUNCTION_ARGS)
{
   PACKED_GRAPH *graph1 = PG_GETARG_GRAPH(0);
   PACKED_GRAPH *graph2 = PG_GETARG_GRAPH(1);
   PACKED_GRAPH *graph;

   graph = PackedGraph_AddGraph(graph1,graph2);
   PG_RETURN_POINTER(graph);
}


Datum graph_has_self_loop(PG_FUNCTION_ARGS)
{
   PACKED_GRAPH *graph = PG_GETARG_GRAPH(0);
   
   PG_RETURN_BOOL(has_self_loop(graph));
}


Datum graph_breadth_first_search(PG_FUNCTION_ARGS)
{
   PACKED_GRAPH *graph = PG_GETARG_GRAPH(0);
   char *node = GET_STR(PG_GETARG_TEXT_P(1));

   PG_RETURN_POINTER(breadth_first_search(graph,node));
}

Datum graph_depth_first_search(PG_FUNCTION_ARGS)
{
   PACKED_GRAPH *graph = PG_GETARG_GRAPH(0);

   PG_RETURN_POINTER(depth_first_search(graph));
}


Datum graph_out_edges(PG_FUNCTION_ARGS)
{
   PACKED_GRAPH *graph = PG_GETARG_GRAPH(0);
   char *node = GET_STR(PG_GETARG_TEXT_P(1));

   PG_RETURN_POINTER(out_edges(graph,node));
}


Datum graph_in_edges(PG_FUNCTION_ARGS)
{
   PACKED_GRAPH *graph = PG_GETARG_GRAPH(0);
   char *node = GET_STR(PG_GETARG_TEXT_P(1));

   PG_RETURN_POINTER(in_edges(graph,node));
}


Datum graph_connected_components(PG_FUNCTION_ARGS)
{
   PACKED_GRAPH *graph = PG_GETARG_GRAPH(0);
   PG_RETURN_POINTER(connected_components(graph));
}
