#include "postgres.h"
#include "utils/array.h"
#include "utils/builtins.h"
#include "executor/executor.h"
#include "catalog/pg_type.h"
#include "fmgr.h"
#include "funcapi.h"
#include "executor/spi.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "graph_op.h"

// declaration of external functions defined in C++ files
extern int surrogate(PACKED_GRAPH *pg);
extern char **_breadth_first_search(PACKED_GRAPH *pg, char *name);
extern char **_depth_first_search(PACKED_GRAPH *pg);
extern char ***_connected_components(PACKED_GRAPH *pg, int* number);

char *AddString( char *string, int *capacity, int *length, char *data, int preserved )
{
   int len, new_capacity, new_length;

   new_capacity = *capacity;
   new_length   = *length;
   len = strlen(data);

   if ( new_length+len+1>new_capacity )
   {
      new_capacity = new_capacity * 2;
      string = repalloc( string, new_capacity );
   }
   memcpy(string+new_length,data,len);
   new_length += len;
   if ( preserved )
   {
      string[new_length] = 0;
      new_length++;
   }
   string[new_length] = 0;
   *length = new_length;
   *capacity = new_capacity;
   return string;
}


UNPACKED_GRAPH *UnpackedGraph_Create(void)
{
   UNPACKED_GRAPH *graph;

   graph = (UNPACKED_GRAPH *) palloc( sizeof(UNPACKED_GRAPH) );
   
   graph->node_capacity = INITIAL_NODE_CAPACITY;
   graph->node_count = 0;
   graph->node_memory_size = 0;
   graph->nodes = (char **) palloc(INITIAL_NODE_CAPACITY * sizeof(char *));   

   graph->edge_capacity = INITIAL_EDGE_CAPACITY;
   graph->edge_count = 0;
   graph->edges = (UNPACKED_GRAPH_EDGE *) palloc(INITIAL_EDGE_CAPACITY*sizeof(UNPACKED_GRAPH_EDGE));

   graph->label_capacity = INITIAL_LABEL_CAPACITY;
   graph->label_count = 0;
   graph->label_memory_size = 0;
   graph->labels = (UNPACKED_GRAPH_LABEL *) palloc(INITIAL_LABEL_CAPACITY*sizeof(UNPACKED_GRAPH_LABEL)); 

   return graph;
}

UNPACKED_GRAPH *UnpackedGraph_Copy( UNPACKED_GRAPH *graph )
{
    UNPACKED_GRAPH *graph1;
    int i;

    graph1 = (UNPACKED_GRAPH *) malloc(sizeof(UNPACKED_GRAPH));
    memcpy(graph1,graph,sizeof(UNPACKED_GRAPH));
   
    // duplicate node information
    graph1->nodes = (char **) malloc( graph->node_capacity*sizeof(char*));
    for(i=0;i<graph->node_count;i++)
       graph1->nodes[i] = pstrdup(graph->nodes[i]);    

    // duplicate edge information    
    graph1->edges = (UNPACKED_GRAPH_EDGE *) malloc(graph->edge_capacity*sizeof(UNPACKED_GRAPH_EDGE));
    memcpy(graph1->edges, graph->edges, graph->edge_count * sizeof(UNPACKED_GRAPH_EDGE));

    // duplicate label information
    graph1->labels = (UNPACKED_GRAPH_LABEL *) malloc(graph->label_capacity * sizeof(UNPACKED_GRAPH_LABEL));
    for(i=0;i<graph->label_count;i++)
    {
       graph1->labels[i].edge = graph->labels[i].edge;
       graph1->labels[i].description = pstrdup(graph->labels[i].description);
    }

    return graph1;
}

void UnpackedGraph_Free( UNPACKED_GRAPH *graph )
{
   int i;

   if ( IS_UNPACKED_GRAPH_INVALID(graph) )
     return;

   for(i=0;i<graph->node_count;i++)
     pfree( graph->nodes[i] );

   for(i=0;i<graph->label_count;i++)
     pfree( graph->labels[i].description );

   pfree(graph->nodes);
   pfree(graph->edges);
   pfree(graph->labels);
   pfree(graph);                
}

int UnpackedGraph_FindNode( UNPACKED_GRAPH *graph, char *nodename )
{
   int i;

   if ( IS_UNPACKED_GRAPH_INVALID(graph) )	
	return -1;

   for(i=0;i<graph->node_count;i++)
     	if ( strcmp(graph->nodes[i], nodename)==0 )  return i;

   return -1;   
}


int UnpackedGraph_InsertNode( UNPACKED_GRAPH *graph, char *nodename )
{
   int i;

   if ( IS_UNPACKED_GRAPH_INVALID(graph) )
	return -1;

   if ( (i=UnpackedGraph_FindNode( graph, nodename ))!=-1 )
	return i;

   if ( graph->node_count==graph->node_capacity )
   {
      i = graph->node_capacity;
      graph->node_capacity = i+i;
      graph->nodes = (char **) repalloc(graph->nodes, sizeof(char *) * (i+i) );
   }

   i = graph->node_count++;
   graph->nodes[i] = pstrdup( nodename );
   graph->node_memory_size += ( strlen(nodename)+1 );

   return i;
}

/*
void UnpackedGraph_RemoveNode( UNPACKED_GRAPH *graph, char *nodename )
{
   int i;

   // not yet done
   if ( IS_UNPACKED_GRAPH_INVALID(graph) )
        return;

   if ( (i=UnpackedGraph_FindNode( graph, nodename ))==-1 )
	return;

   // remove edges and labels which point to those edges, that contain "i"
}
*/

/*
void UnpackedGraph_RemoveLabel_ByEdge( UNPACKED_GRAPH *graph, int edge_no, char *description)
{
   UNPACKED_GRAPH_LABEL *label;
   int i, j;
 
   if ( (i=UnpackedGraph_FindLabel_ByEdge( graph, edge_no, description ))!=-1 )
   {
        for(j=i+1;j<graph->label_count;j++)
        {
	     memcpy(label,label+1,sizeof(UNPACKED_GRAPH_LABEL));
	     label++;
        }
	graph->label_count--;
   }
}

void UnpackedGraph_RemoveLabel_ByName( UNPACKED_GRAPH *graph, char *src, char *dest, char *label_description)
{
   int edge_no;
   UNPACKED_GRAPH_EDGE *edge;

   if ( (edge_no=UnpackedGraph_FindEdge_ByName( graph, src, dest))!=-1 )
   {
	UnpackedGraph_RemoveLabel_ByEdge( graph, edge_no, label_description );
   }
}

void UnpackedGraph_RemoveEdge_ByName( UNPACKED_GRAPH *graph, char *src, char *dest, char *label_description )
{
   UNPACKED_GRAPH_EDGE *edge;
   UNPACKED_GRAPH_LABEL *label;
   int i, j, edge_no;

   if ( (edge_no=UnpackedGraph_FindEdge_ByName( graph, src, dest ))!=-1 )
   {
       // we first have to remove labels which are attached to this edge...
       // to-do
     
       // we can now safely delete this edge
       edge = &(graph->edges[i]);
       for(i=edge_no+1;i<graph->edge_count;i++)
       {
	   memcpy(edge,edge+1,sizeof(UNPACKED_GRAPH_EDGE));
	   edge++;
       }
       graph->edge_count--;
   }
}
*/

int UnpackedGraph_FindEdge( UNPACKED_GRAPH *graph, int src, int dest )
{
   UNPACKED_GRAPH_EDGE *edge;
   int i;

   if ( IS_UNPACKED_GRAPH_INVALID(graph) )
	return -1;

   for(i=0,edge=graph->edges;i<graph->edge_count;i++,edge++)
	if ( (edge->src==src) && (edge->dest==dest) )  
		return i;

   return -1;
}

int UnpackedGraph_FindEdge_ByName( UNPACKED_GRAPH *graph, char *src, char *dest )
{
   int n1, n2;

   n1 = UnpackedGraph_FindNode(graph,src);
   n2 = UnpackedGraph_FindNode(graph,dest);
   if ( (n1==-1) || (n2==-1) )	
	return -1;

   return UnpackedGraph_FindEdge( graph, n1, n2 );
}

int UnpackedGraph_InsertEdge_ById( UNPACKED_GRAPH *graph, int src, int dest )
{
   UNPACKED_GRAPH_EDGE *edge;
   int i, edge_no;

   if ( IS_UNPACKED_GRAPH_INVALID(graph) )
        return -1;

   if ( (edge_no=UnpackedGraph_FindEdge( graph, src, dest ))!=-1 )
   {
        return edge_no;
   }

   if ( graph->edge_count==graph->edge_capacity )
   {
      graph->edge_capacity = graph->edge_capacity + graph->edge_capacity;
      graph->edges = (UNPACKED_GRAPH_EDGE *) repalloc(graph->edges, sizeof(UNPACKED_GRAPH_EDGE) * graph->edge_capacity );
   }
   
   i = graph->edge_count;
   edge = &(graph->edges[i]);
   graph->edge_count++;
   edge->src = src;
   edge->dest = dest;
  
   return i;
}

int UnpackedGraph_FindLabel_ByEdge( UNPACKED_GRAPH *graph, int edge_no, char *label_description )
{
   UNPACKED_GRAPH_LABEL *label;
   int i;

   if ( IS_UNPACKED_GRAPH_INVALID(graph) )
	return -1;

   for(i=0, label=graph->labels;i<graph->label_count;i++,label++)
   {
      if ( (label->edge==edge_no) && !strcmp(label->description,label_description) )
	  return i;
   }

   return -1;
}

int UnpackedGraph_InsertEdge_ByName( UNPACKED_GRAPH *graph, char *src, char *dest )
{
   int s, d;

   s = UnpackedGraph_InsertNode(graph,src);
   d = UnpackedGraph_InsertNode(graph,dest);

   return UnpackedGraph_InsertEdge_ById( graph, s, d );
}

int UnpackedGraph_InsertEdge_BySpec( UNPACKED_GRAPH *graph, char *spec, int insert_mode )
{
   char *p;
   int ret, src, dest;

   ret = -1;
   if ( (p=strtok(spec,", "))!=NULL )
   {
       if ( insert_mode )
          src = UnpackedGraph_InsertNode( graph, p );
       else
       {
          if ( (src=UnpackedGraph_FindNode( graph, p ))==-1 )
             elog( ERROR, "node %s does not exist", p);
       }	 

       if ( (p=strtok(NULL,", "))!=NULL )
       {
          if ( insert_mode )
             dest = UnpackedGraph_InsertNode( graph, p );
          else
          {
             if ( (dest=UnpackedGraph_FindNode( graph, p ))==-1 )
                elog( ERROR, "node %s does not exist", p);
          }
          ret = UnpackedGraph_InsertEdge_ById( graph, src, dest );
       }
   }

   return ret;
}

int UnpackedGraph_FindLabel_ById( UNPACKED_GRAPH *graph, int src, int dest, char *label )
{
   int i, edge_no;

   if ( IS_UNPACKED_GRAPH_INVALID(graph) )
        return  -1;

   if ( (edge_no=UnpackedGraph_FindEdge( graph, src, dest ))==-1 )
 	return  -1;
   
   for(i=0;i<graph->label_count;i++)
	if ( (graph->labels[i].edge==edge_no) && (!strcmp(graph->labels[i].description,label)) )	
		return i;;
   return -1;
}

int UnpackedGraph_InsertLabel_ByEdge( UNPACKED_GRAPH *graph, int edge_no, char *label )
{
   UNPACKED_GRAPH_LABEL *p;
   int i;

   if ( IS_UNPACKED_GRAPH_INVALID(graph) )
	return -1;

   if ( (i=UnpackedGraph_FindLabel_ByEdge( graph, edge_no, label ))!=-1 )
   {
	return i;
   }

   if ( graph->label_count==graph->label_capacity )
   {
        graph->label_capacity = graph->label_capacity + graph->label_capacity;
        graph->labels = (UNPACKED_GRAPH_LABEL *) repalloc(graph->labels, sizeof(UNPACKED_GRAPH_LABEL) * graph->label_capacity );
   }

   i = graph->label_count;
   p = &(graph->labels[i]);
   graph->label_count++;
   p->edge = edge_no;
   p->description = pstrdup( label );
   graph->label_memory_size += ( strlen(label)+1 );

   return i;
}

int UnpackedGraph_InsertLabel_ById( UNPACKED_GRAPH *graph, int src, int dest, char *label )
{
   int edge_no;

   if ( IS_UNPACKED_GRAPH_INVALID(graph) )
        return -1;

   if ( (edge_no=UnpackedGraph_FindEdge( graph, src, dest ))==-1 )
        edge_no = UnpackedGraph_InsertEdge_ById( graph, src, dest );

   return UnpackedGraph_InsertLabel_ByEdge( graph, edge_no, label );
}

int UnpackedGraph_InsertLabel_ByName( UNPACKED_GRAPH *graph, char *src, char *dest, char *label )
{
   int s, d, edge_no;

   s = UnpackedGraph_InsertNode(graph,src);
   d = UnpackedGraph_InsertNode(graph,dest);

   edge_no = UnpackedGraph_InsertEdge_ById( graph, s, d);
   return UnpackedGraph_InsertLabel_ByEdge( graph, edge_no, label );
}

int UnpackedGraph_InsertLabel_BySpec( UNPACKED_GRAPH *graph, char *spec, int insert_mode )
{
   char *p, *q;
   int ret, edge;

   ret = -1;
   if ( (p=strtok(spec,", "))!=NULL )
   {
       if ( (q=strtok(NULL,", "))!=NULL )
       {
           if ( insert_mode )
              edge = UnpackedGraph_InsertEdge_ByName( graph, p, q );
           else
           {
              if ( (edge=UnpackedGraph_FindEdge_ByName( graph, p, q ))==-1 )
	         elog( ERROR, "edge (%s,%s) does not exist", p, q);
           }
           
           if ( (p=strtok(NULL,","))!=NULL )
	   {
              q = p + strlen(p)-1;
	      while (*q==' ') *q--='\0';

              ret = UnpackedGraph_InsertLabel_ByEdge( graph, edge, p);
	   }
           else
              elog( ERROR, "label format is illegal" );
       }
   }

   return ret;
}


void UnpackedGraph_Print( UNPACKED_GRAPH *graph )
{
   int i, edge_no;
   printf( "node [%d/%d], edge [%d/%d], label [%d/%d]\n", 
           graph->node_count, graph->node_capacity,
           graph->edge_count, graph->edge_capacity,
           graph->label_count, graph->label_capacity );
   printf( "node_memory_size=%d, label_memory_size=%d\n", 
           graph->node_memory_size, graph->label_memory_size );

   printf( "-nodes...  \n");
   for(i=0;i<graph->node_count;i++)
      printf("[%s]", graph->nodes[i] );
   printf( "\n-edges...\n  ");

   for(i=0;i<graph->edge_count;i++)
      printf("[%d,%d]", graph->edges[i].src, graph->edges[i].dest );
   printf("\n");

   printf( "-labels...\n  ");
   for(i=0;i<graph->label_count;i++)
   {
      edge_no = graph->labels[i].edge;
      printf("[%d,%d](%s)", graph->edges[edge_no].src, graph->edges[edge_no].dest, graph->labels[i].description );
   }
   printf("\n");
   
}

PACKED_GRAPH *Pack( UNPACKED_GRAPH *graph )
{
   PACKED_GRAPH *ret;
   GRAPH_OFFSET offsets[ID_MAX_OFFSET], offset;
   UNPACKED_GRAPH_EDGE *edge;
   UNPACKED_GRAPH_LABEL *label;
   char *ptr, *src, *dest;
   int i, length, len, edge_no;

   offsets[ID_NODE_TABLE_OFFSET] = 9*sizeof(uint32);
   offsets[ID_LABEL_TABLE_OFFSET] = graph->node_count * sizeof(GRAPH_OFFSET);
   offsets[ID_NODES_OFFSET] = graph->label_count * sizeof(GRAPH_OFFSET);
   offsets[ID_EDGES_OFFSET] = graph->node_memory_size;
   offsets[ID_LABELS_OFFSET] = graph->edge_count * 2 * sizeof(GRAPH_OFFSET);
   for(i=1;i<ID_MAX_OFFSET;i++)
	offsets[i] += offsets[i-1];

   length = offsets[ID_LAST_OFFSET] + graph->label_memory_size + 2*sizeof(GRAPH_OFFSET)*graph->label_count;
  
   ret = (PACKED_GRAPH *) palloc( length ); 
   ret->size = length;
   ret->nnodes = graph->node_count;
   ret->nedges = graph->edge_count;
   ret->nlabels = graph->label_count;
   memcpy( ret->offsets, offsets, sizeof(GRAPH_OFFSET)*ID_MAX_OFFSET );
   
   src = ret->data;
   dest = ret->data + sizeof(GRAPH_OFFSET) * ( graph->node_count+graph->label_count );
   offset = 0;

   // write out node offsets and nodes information
   for(i=0;i<graph->node_count;i++)
   {
	ptr = graph->nodes[i];
     	len = strlen(ptr)+1;
        *((GRAPH_OFFSET *) src) = offset;
	memcpy( dest, ptr, len );
	src += sizeof(GRAPH_OFFSET);
	dest += len;
	offset += len;
   }

   // write out edge information (there is no edge offset)
   for(i=0, edge=graph->edges;i<graph->edge_count;i++,edge++)
   {
        *((GRAPH_OFFSET *) (dest)) = edge->src;
        *((GRAPH_OFFSET *) (dest+sizeof(GRAPH_OFFSET))) = edge->dest;
	dest += (2*sizeof(GRAPH_OFFSET));
   }

   // write out label offsets and label information
   offset = 0;
   for(i=0, label=graph->labels;i<graph->label_count;i++,label++)
   {
        edge_no = label->edge;
        *((GRAPH_OFFSET *) src) = offset;
	*((GRAPH_OFFSET *) dest) = graph->edges[edge_no].src;
        *((GRAPH_OFFSET *) (dest+sizeof(GRAPH_OFFSET))) = graph->edges[edge_no].dest;
   	ptr = label->description;
        len = strlen(ptr)+1;
        memcpy( dest+2*sizeof(GRAPH_OFFSET), ptr, len );
	offset += (len+2*sizeof(GRAPH_OFFSET));
        src += sizeof(GRAPH_OFFSET);
        dest += ( 2*sizeof(GRAPH_OFFSET)+len );
   }


   return ret;
}

UNPACKED_GRAPH *Unpack( PACKED_GRAPH *graph )
{
   UNPACKED_GRAPH *g;
   GRAPH_OFFSET *offset;
   char *src;
   int length, len, v, i, n1, n2;

   //test
   surrogate(graph);
    
   g = (UNPACKED_GRAPH *) palloc( sizeof(UNPACKED_GRAPH) );

   v = PACKED_GRAPH_NUM_NODES(graph);
   g->node_capacity = v+10;
   g->node_count = v;
   g->nodes = (char **) palloc( g->node_capacity * sizeof(char *));
   length = 0;

   src = PACKED_GRAPH_NODES(graph);
   for(i=0;i<v;i++)
   {
      len = strlen(src)+1;
      g->nodes[i] = pstrdup(src);
      length += len;
      src += len;
   }
   g->node_memory_size = length;


   v = PACKED_GRAPH_NUM_EDGES(graph);
   g->edge_capacity = v+10;
   g->edge_count = v;
   g->edges = (UNPACKED_GRAPH_EDGE *) palloc(g->edge_capacity*sizeof(UNPACKED_GRAPH_EDGE));
   offset = (GRAPH_OFFSET *) PACKED_GRAPH_EDGES(graph);
   for(i=0;i<v;i++)
   {
      g->edges[i].src = *offset;
      g->edges[i].dest = *(offset+1);
      offset += 2;
   }

   v = PACKED_GRAPH_NUM_LABELS(graph);
   g->label_capacity = v+10;
   g->label_count = 0;
   g->labels = (UNPACKED_GRAPH_LABEL *) palloc( g->label_capacity * sizeof(UNPACKED_GRAPH_LABEL));
   g->label_memory_size = 0;


   src = PACKED_GRAPH_LABELS(graph);
   for(i=0;i<v;i++)
   {
      offset = (GRAPH_OFFSET *) src;
      n1 = *offset;        
      n2 = *(offset+1);
      src += (2*sizeof(GRAPH_OFFSET));
      UnpackedGraph_InsertLabel_ById(g, n1, n2, src);
      len = strlen(src);
      src += (len+1);
   }

   return g;
}

int PackedGraph_FindNode( PACKED_GRAPH *graph, char *nodename )
{
   GRAPH_OFFSET *offset;
   char *nodes;
   int i;

   nodes = PACKED_GRAPH_NODES(graph);
   offset = (GRAPH_OFFSET *) PACKED_GRAPH_NODE_TABLE(graph);

   for(i=0;i<graph->nnodes;i++)
        if ( strcmp(nodes+offset[i],nodename)==0 ) return i;

   return -1;
}

int PackedGraph_FindEdge_ByName( PACKED_GRAPH *graph, char *src, char *dest )
{
   GRAPH_OFFSET *offset;
   int i, src_no, dest_no;

   if ( (src_no=PackedGraph_FindNode(graph, src))==-1 )
 	return -1;
   if ( (dest_no=PackedGraph_FindNode(graph,dest))==-1 )
	return -1;
   
   offset = (GRAPH_OFFSET *) PACKED_GRAPH_EDGES(graph);
   for(i=0;i<graph->nnodes;i++,offset+=2)
        if ( (offset[0]==src_no) && (offset[1]==dest_no) )  return i;
        
   return -1;
}

PACKED_GRAPH *PackedGraph_InsertNode_V( PACKED_GRAPH *graph, char **value )
{
   UNPACKED_GRAPH *ug;
   char **p;
   ug = Unpack(graph);

   p = value;
   while ( *p!=NULL )
   {
      UnpackedGraph_InsertNode( ug, *p );
      p++;
   }

   return Pack(ug);
}

PACKED_GRAPH *PackedGraph_InsertNode( PACKED_GRAPH *graph, char *value )
{
   PACKED_GRAPH *pg;
   char **p, *q;
   int i, j;

   i=0;
   q=value-1;
   while (1)
   {
      q = strchr(q+1,',');
      if (q==NULL) break;
      i++;
   }

   p = (char **) palloc( sizeof(char *) * (i+2) );

   q = strtok( value, ", " );
   p[0] = q;
   for(j=1;j<=(i+1);j++)
   {
      q = strtok( NULL, ", " );
      p[j] = q;      
   }
   p[i+1] = NULL;
   
   pg = PackedGraph_InsertNode_V( graph, p );
   pfree(p);
   return pg;   
}

PACKED_GRAPH *PackedGraph_InsertEdge_V( PACKED_GRAPH *graph, char **value )
{
   UNPACKED_GRAPH *ug;
   char **p;
   ug = Unpack(graph);

   p = value;
   while ( *p!=NULL )
   {
      UnpackedGraph_InsertEdge_BySpec( ug, *p, 1 );
      p++;
   }

   return Pack(ug);
}

PACKED_GRAPH *PackedGraph_InsertEdge( PACKED_GRAPH *graph, char *value )
{
   PACKED_GRAPH *pg;
   char **p, *q, *s, *r;
   int capacity, counter;

   capacity = 10;
   counter = 0;
   p = (char **) palloc( sizeof(char *) * capacity );

   s = value;
   while (1)
   {
     if ( (q=strchr(s,'('))!=NULL )
     {
        if ( (r=strchr(q+1,')'))!=NULL )
        { 
           
	    if ( counter==capacity )
            {
                capacity = capacity + capacity;
                p = (char **) repalloc( p, sizeof(char *) * capacity );
            }
            *r = 0;
            p[counter++] = q+1;
            s = r+1;
        }
        else  
           elog( ERROR, "unmatching quote" );
     }
     else  break;
   }

   p[counter] = NULL;

   pg = PackedGraph_InsertEdge_V( graph, p );
   pfree(p);
   return pg;   
}

PACKED_GRAPH *PackedGraph_InsertLabel_V( PACKED_GRAPH *graph, char **value )
{
   UNPACKED_GRAPH *ug;
   char **p;
   ug = Unpack(graph);

   p = value;
   while ( *p!=NULL )
   {
      UnpackedGraph_InsertLabel_BySpec( ug, *p, 1 );
      p++;
   }

   return Pack(ug);
}

PACKED_GRAPH *PackedGraph_InsertLabel( PACKED_GRAPH *graph, char *value )
{
   PACKED_GRAPH *pg;
   char **p, *q, *s, *r;
   int capacity, counter;

   capacity = 10;
   counter = 0;
   p = (char **) palloc( sizeof(char *) * capacity );

   s = value;
   while (1)
   {
     if ( (q=strchr(s,'('))!=NULL )
     {
        if ( (r=strchr(q+1,')'))!=NULL )
        {

            if ( counter==capacity )
            {
                capacity = capacity + capacity;
                p = (char **) repalloc( p, sizeof(char *) * capacity );
            }
            *r = 0;
            p[counter++] = q+1;
            s = r+1;
        }
        else
           elog( ERROR, "unmatching quote" );
     }
     else  break;
   }

   p[counter] = NULL;

   pg = PackedGraph_InsertLabel_V( graph, p );
   pfree(p);
   return pg;
}

PACKED_GRAPH *PackedGraph_AddGraph( PACKED_GRAPH *graph1, PACKED_GRAPH *graph2 )
{
   UNPACKED_GRAPH *old, *op1, *op2;
   UNPACKED_GRAPH_LABEL *labels;
   char *src, *dest;
   int i, j;

   old = Unpack(graph1);
   op1 = UnpackedGraph_Copy(old);
   op2 = Unpack(graph2);


   // insert nodes
   for(i=0;i<op2->node_count;i++)
   {
      UnpackedGraph_InsertNode(op1, op2->nodes[i]);
   }

   // insert edges
   for(i=0;i<op2->edge_count;i++)
   {
      src  = op2->nodes[op2->edges[i].src];
      dest = op2->nodes[op2->edges[i].dest];

      UnpackedGraph_InsertEdge_ByName(op1,src,dest);
   }

   for(i=0;i<op2->label_count;i++)
   {
      labels = &(op2->labels[i]);
      j = labels->edge;
      src  = op2->nodes[op2->edges[i].src];
      dest = op2->nodes[op2->edges[i].dest];

      UnpackedGraph_InsertLabel_ByName(op1, src, dest, labels->description);
   }

   return Pack(op1);
}

// operations
bool has_self_loop(PACKED_GRAPH* g)
{
   GRAPH_OFFSET *offset;
   int v, i, n1, n2;

   v = PACKED_GRAPH_NUM_EDGES(g);
   offset = (GRAPH_OFFSET *) PACKED_GRAPH_EDGES(g);
   for(i=0;i<v;i++)
   {
      n1 = (int) *offset;
      n2 = (int) *(offset+1);
      if ( n1==n2 ) return true;
      offset += 2;
   }
   return false;
}



// the following is for other things
#define NDIM 1

static uint textoid = 0;

int calculate_string_size( char **value, int *num )
{
   int sum, i, length;
   char **p;

   p = value;
   sum =0;
   i=0;
   while ( *p!=NULL )
   {
      length = strlen((char *) (*p)) + VARHDRSZ;
      length = ((length+3) / 4 )*4;
      sum = sum + length;
      p++;
      i++;
   }

   *num = i;   
   return sum;
}

ArrayType *text_array_create( char **value )
{
   ArrayType *r;
   SPITupleTable *spi_tuptable;
   TupleDesc spi_tupdesc;
   HeapTuple spi_tuple;
   int nbytes, ret, proc, num, length;
   char **val;
   char *p;

   nbytes = ARR_OVERHEAD(NDIM) + calculate_string_size(value, &num);

   r = (ArrayType *) palloc0(nbytes);

   ARR_SIZE(r) = nbytes;
   ARR_NDIM(r) = NDIM;

   if ( textoid==0 )
   {

      if ((ret=SPI_connect())<0)
        elog( ERROR, "new_textArray: SPI_connect returns %d",ret);

      ret = SPI_execute("select oid from pg_type where typname='text'",true,0);
      proc = SPI_processed;
      if ((ret==SPI_OK_SELECT) && (proc>0))
      {
         spi_tuptable = SPI_tuptable;
         spi_tupdesc = spi_tuptable->tupdesc;
         spi_tuple = spi_tuptable->vals[0];
         p = SPI_getvalue(spi_tuple, spi_tupdesc,1);
         textoid = atoi(p);
      }
      else
        elog(ERROR, "unable to find oid for text datatype!" );

      SPI_finish();
   }

   ARR_ELEMTYPE(r) = textoid;

   
   *((int *)ARR_DIMS(r)) = num;
   *((int *)ARR_LBOUND(r)) = 1;

   p = (char *) ARR_DATA_PTR(r);
      
   val = value;
   while ( *val!=NULL )
   {
      length = strlen( (char *) (*val));
      VARATT_SIZEP((text *)p) = VARHDRSZ+length;
      memcpy((void*) VARDATA((text *)p), (char *) (*val), length );

      
      length = ((length + VARHDRSZ + 3) / 4 ) * 4;
      p = p+ length;

      val = val + 1;
   }

   return r;
}

ArrayType* breadth_first_search(PACKED_GRAPH *g, char *node)
{
   ArrayType *v;
   char **p;

   p = _breadth_first_search(g, node);
   v = text_array_create(p);
   return v;
}

ArrayType* depth_first_search(PACKED_GRAPH *g)
{
   ArrayType *v;
   char **p;

   p = _depth_first_search(g);
   v = text_array_create(p);
   return v;
}

ArrayType* connected_components(PACKED_GRAPH *g)
{
   ArrayType *r;
   UNPACKED_GRAPH **ug, *orig_graph;
   PACKED_GRAPH   **pg;
   char ***cc, **level1, *level2;
   int num, nbytes, result, proc, length;
   int i, j, k, w;
   SPITupleTable *spi_tuptable;
   TupleDesc spi_tupdesc;
   HeapTuple spi_tuple;
   char *p;
   uint graphoid = 0;
   GRAPH_OFFSET *offset;
   char *pivot;
   int src, dest;

   // run boost::connect_components algorithm, [OUT] num is the total number
   // of connected components
   cc = _connected_components(g, &num);

   orig_graph = Unpack(g);
   ug = (UNPACKED_GRAPH **) palloc( sizeof(UNPACKED_GRAPH *) * num );
   pg = (PACKED_GRAPH **) palloc( sizeof(PACKED_GRAPH*) * num);

   // split nodes in original graph into different graphs
   i = 0;
   while ( *cc != NULL )
   {
      ug[i] = (UNPACKED_GRAPH *) UnpackedGraph_Create();
      level1 = *cc;
      while ( *level1 != NULL )
      {
         level2 = *level1;
         UnpackedGraph_InsertNode( ug[i], level2 );         
         level1++;
      }            
      cc++;
      i++;
   }

   // iterate through edges in original graph and find which connected
   // component this edge belong to

   w = -1;
   i = PACKED_GRAPH_NUM_EDGES(g);   
   offset = (GRAPH_OFFSET *) PACKED_GRAPH_EDGES(g);
   for(j=0;j<i;j++)
   {
       src = (int) *offset;
       dest = (int) *(offset+1);

       // look for where this edge is located in the split graph
       for(k=0;k<num;k++)
          if ( (w=UnpackedGraph_FindNode(ug[k], orig_graph->nodes[src]))!=-1 ) break;

       UnpackedGraph_InsertEdge_ByName(ug[k],orig_graph->nodes[src], orig_graph->nodes[dest]);
       offset +=2;            
   }

   // now, we have to add labels (from original PACKED_GRAPH)
   i = PACKED_GRAPH_NUM_LABELS(g);
   pivot = PACKED_GRAPH_LABELS(g);
   for(j=0;j<i;j++)
   {
      offset = (GRAPH_OFFSET *) pivot;
      src = *offset;
      dest = *(offset+1);
      pivot = pivot + 2*sizeof(GRAPH_OFFSET);
      
      // look for where this edge is located in the split graph
      for(k=0;k<num;k++)
         if ( (w=UnpackedGraph_FindNode(ug[k], orig_graph->nodes[src]))!=-1 ) 
             break;

       UnpackedGraph_InsertLabel_ByName(ug[k],
                       orig_graph->nodes[src],
                       orig_graph->nodes[dest],pivot);

      pivot = pivot + strlen((char*) pivot) + 1;  
   }
  
   
   nbytes = ARR_OVERHEAD(NDIM);   
   for(i=0;i<num;i++)
   {
      pg[i] = Pack(ug[i]);
      UnpackedGraph_Free(ug[i]);
      nbytes = nbytes + (( pg[i]->size + 3)/4)*4;
   }

   pfree( ug );
   
   // create Graph Array  

   r = (ArrayType *) palloc0(nbytes);

   ARR_SIZE(r) = nbytes;
   ARR_NDIM(r) = NDIM;

   // look up oid for graph datatype
   if ((result=SPI_connect())<0)
       elog( ERROR, "connected_components: SPI_connect returns %d",result);

   result = SPI_execute("select oid from pg_type where typname='graph'",true,0);
   proc = SPI_processed;
   if ((result==SPI_OK_SELECT) && (proc>0))
   {
      spi_tuptable = SPI_tuptable;
      spi_tupdesc = spi_tuptable->tupdesc;
      spi_tuple = spi_tuptable->vals[0];
      p = SPI_getvalue(spi_tuple, spi_tupdesc,1);
      graphoid = atoi(p);
   }
   else
     elog(ERROR, "unable to find oid for graph datatype!" );

   SPI_finish();

   ARR_ELEMTYPE(r) = graphoid;

   *((int *)ARR_DIMS(r)) = num;
   *((int *)ARR_LBOUND(r)) = 1;

   p = (char *) ARR_DATA_PTR(r);   
   for(i=0;i<num;i++)
   {
      length = ((pg[i]->size + 3)/4)*4;
      
      memcpy( (void*) p, (void*) pg[i], pg[i]->size);
      p = p + length;
   }     

   for(i=0;i<num;i++)
   {
      pfree(pg[i]);
   }

   return r;
}


ArrayType* out_edges(PACKED_GRAPH *g, char *node)
{
   UNPACKED_GRAPH *ug;
   int node_id, i, index, nedges;
   char **list;
   char buffer[2*2048];

   ug = Unpack(g);
   node_id = UnpackedGraph_FindNode( ug, node );
   
   nedges = ug->edge_count;
   list = (char **) palloc(sizeof(char *) *(nedges+1));
   index = 0;
   if ( node_id!=-1 )
   {
      for(i=0;i<nedges;i++)
      {
         if (ug->edges[i].src==node_id) 
         {
            sprintf(buffer,"(%s,%s)", ug->nodes[node_id], ug->nodes[ug->edges[i].dest]);
            list[index++] = (char *) pstrdup(buffer);
         }
      }
   }   
   list[index] = NULL;

   UnpackedGraph_Free(ug);
   return text_array_create(list);
}


ArrayType* in_edges(PACKED_GRAPH *g, char *node)
{
   UNPACKED_GRAPH *ug;
   int node_id, i, index, nedges;
   char **list;
   char buffer[2*2048];

   ug = Unpack(g);
   node_id = UnpackedGraph_FindNode( ug, node );

   nedges = ug->edge_count;
   list = (char **) palloc(sizeof(char *) *(nedges+1));
   index = 0;
   if ( node_id!=-1 )
   {
      for(i=0;i<nedges;i++)
      {
         if (ug->edges[i].dest==node_id)
         {
            sprintf(buffer,"(%s,%s)", ug->nodes[ug->edges[i].src], ug->nodes[node_id]);
            list[index++] = (char *) pstrdup(buffer);
         }
      }
   }
   list[index] = NULL;
   UnpackedGraph_Free(ug);
   return text_array_create(list);
}

