/*
**  PostGraph -- Graph data management in PostgreSQL
**
**  Copyright (C) 2005, 2006  Ruey-Lung Hsiao,  D.S. Parker
**  (rlhsiao@cs.ucla.edu, stott@cs.ucla.edu)
**
**  UCLA Center for Computational Biology
**  635 Charles Young Drive South, Suite 225
**  Los Angeles, CA 90095-7332  USA
**
**  UCLA Computer Science Dept.
**  3532 Boelter Hall
**  Los Angeles, CA 90095-1596  USA
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public License
**  as published by the Free Software Foundation; either version 2.1 of
**  the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
**  See the GNU Lesser General Public License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the
**  Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
**  Boston, MA  02110-1301  USA
*/

#pragma GCC system_header

extern "C" {

#ifndef __GRAPH_OP_HPP
#define __GRAPH_OP_HPP

typedef int GRAPH_OFFSET;		/*  GRAPH OFFSET is used to keep track of offsets of object in postgresql 
                                         *  graph representation */
#define INVALID_GRAPH_OFFSET  -1
#define IS_GRAPH_OFFSET_VALID(o)	((o)!=INVALID_GRAPH_OFFSET)
#define IS_GRAPH_OFFSET_INVALID(o)	((o)==INVALID_GRAPH_OFFSET)
#define IS_UNPACKED_GRAPH_VALID(g)	((g)!=NULL)
#define IS_UNPACKED_GRAPH_INVALID(g)	((g)==NULL)

#define GET_STR(textp) DatumGetCString(DirectFunctionCall1(textout, PointerGetDatum(textp)))

#define INITIAL_NODE_CAPACITY           10
#define INITIAL_EDGE_CAPACITY           10
#define INITIAL_LABEL_CAPACITY          10
typedef struct UNPACKED_GRAPH_EDGE
{
  int src, dest;
} UNPACKED_GRAPH_EDGE;

typedef struct UNPACKED_GRAPH_LABEL
{
  int edge;
  char *description;
} UNPACKED_GRAPH_LABEL;

typedef struct UNPACKED_GRAPH
{
   int node_capacity, node_count;
   int node_memory_size;
   char **nodes;
   int edge_capacity, edge_count;
   UNPACKED_GRAPH_EDGE *edges;
   int label_capacity, label_count;
   int label_memory_size;
   UNPACKED_GRAPH_LABEL *labels;
} UNPACKED_GRAPH;


#define ID_NODE_TABLE_OFFSET    0
#define ID_LABEL_TABLE_OFFSET  	1
#define ID_NODES_OFFSET		2
#define ID_EDGES_OFFSET		3
#define ID_LABELS_OFFSET	4
#define ID_LAST_OFFSET          4
#define ID_MAX_OFFSET		(ID_LABELS_OFFSET+1)

typedef struct PACKED_GRAPH
{
   unsigned int size;
   unsigned int nnodes, nedges, nlabels;
   GRAPH_OFFSET offsets[ID_MAX_OFFSET];
   char data[1];
} PACKED_GRAPH;


#define PACKED_GRAPH_SIZE(g)		((g)->size)
#define PACKED_GRAPH_NUM_NODES(g)       ((g)->nnodes)
#define PACKED_GRAPH_NUM_EDGES(g)       ((g)->nedges)
#define PACKED_GRAPH_NUM_LABELS(g)      ((g)->nlabels)

#define PACKED_GRAPH_NODE_TABLE(g)	(((char *)(g))+(g)->offsets[ID_NODE_TABLE_OFFSET])
#define PACKED_GRAPH_LABEL_TABLE(g)     (((char *)(g))+(g)->offsets[ID_LABEL_TABLE_OFFSET])
#define PACKED_GRAPH_NODES(g)      	(((char *)(g))+(g)->offsets[ID_NODES_OFFSET])
#define PACKED_GRAPH_EDGES(g)     	(((char *)(g))+(g)->offsets[ID_EDGES_OFFSET])
#define PACKED_GRAPH_LABELS(g)          (((char *)(g))+(g)->offsets[ID_LABELS_OFFSET])

#define PACKED_GRAPH_NODE_TABLE_LENGTH(g)	(((g)->nnodes)*sizeof(GRAPH_OFFSET))
#define PACKED_GRAPH_LABEL_TABLE_LENGTH(g)      (((g)->nlabels)*sizeof(GRAPH_OFFSET))
#define PACKED_GRAPH_NODES_LENGTH(g)       	(((g)->offsets[ID_EDGES_OFFSET])-((g)->offsets[ID_NODES_OFFSET]))
#define PACKED_GRAPH_EDGES_LENGTH(g)		(((g)->nedges)*sizeof(GRAPH_OFFSET)*2)
#define PACKED_GRAPH_LABELS_LENGTH(g)		(((g)->size)-((g)->offsets[ID_LABELS_OFFSET]))

char *AddString( char *string, int *capacity, int *length, char *data, int preserved );
UNPACKED_GRAPH *UnpackedGraph_Create(void);
UNPACKED_GRAPH *UnpackedGraph_Copy( UNPACKED_GRAPH *graph );
void UnpackedGraph_Free( UNPACKED_GRAPH *graph );
int UnpackedGraph_FindNode( UNPACKED_GRAPH *graph, char *nodename );
int UnpackedGraph_InsertNode( UNPACKED_GRAPH *graph, char *nodename );
void UnpackedGraph_RemoveNode( UNPACKED_GRAPH *graph, char *nodename );
int UnpackedGraph_FindEdge( UNPACKED_GRAPH *graph, int src, int dest );
int UnpackedGraph_FindEdge_ByName( UNPACKED_GRAPH *graph, char *src, char *dest );
int UnpackedGraph_InsertEdge_ByName( UNPACKED_GRAPH *graph, char *src, char *dest );
int UnpackedGraph_InsertEdge_ById( UNPACKED_GRAPH *graph, int src, int dest );
int UnpackedGraph_InsertEdge_BySpec( UNPACKED_GRAPH *graph, char *spec, int insert_mode );
int UnpackedGraph_FindLabel_ByEdge( UNPACKED_GRAPH *graph, int edge_no, char *label );
int UnpackedGraph_FindLabel_ById( UNPACKED_GRAPH *graph, int src, int dest, char *label );
int UnpackedGraph_InsertLabel_ByEdge( UNPACKED_GRAPH *graph, int edge_no, char *label );
int UnpackedGraph_InsertLabel_ById( UNPACKED_GRAPH *graph, int src, int dest, char *label );
int UnpackedGraph_InsertLabel_ByName( UNPACKED_GRAPH *graph, char *src, char *dest, char *label );
int UnpackedGraph_InsertLabel_BySpec( UNPACKED_GRAPH *graph, char *spec, int insert_mode );
void UnpackedGraph_Print( UNPACKED_GRAPH *graph );

UNPACKED_GRAPH *Unpack( PACKED_GRAPH *graph );
PACKED_GRAPH *Pack( UNPACKED_GRAPH *graph );
PACKED_GRAPH *PackedGraph_InsertNode_V( PACKED_GRAPH *graph, char **value );
PACKED_GRAPH *PackedGraph_InsertNode( PACKED_GRAPH *graph, char *value );
PACKED_GRAPH *PackedGraph_InsertEdge_V( PACKED_GRAPH *graph, char **value );
PACKED_GRAPH *PackedGraph_InsertEdge( PACKED_GRAPH *graph, char *value );
PACKED_GRAPH *PackedGraph_InsertLabel_V( PACKED_GRAPH *graph, char **value );
PACKED_GRAPH *PackedGraph_InsertLabel( PACKED_GRAPH *graph, char *value );
PACKED_GRAPH *PackedGraph_AddGraph( PACKED_GRAPH *graph1, PACKED_GRAPH *graph2 );


int PackedGraph_FindNode( PACKED_GRAPH *graph, char *nodename );
int PackedGraph_FindEdge_ByName( PACKED_GRAPH *graph, char *src, char *dest );
}
#endif





